/// matrix2d.h
///
/// (c) 2008-2020 by Carsten Busse <carsten.busse@googlemail.com>
///             Bastian Spiegel <bs@tkscript.de> (additional coding)
///     - Distributed under terms of the Lesser GNU General Public License (LGPL).
///       See COPYING and <http://www.gnu.org/licenses/licenses.html#LGPL> for further information.
///
///

#ifndef __MATRIX2D_H_
#define __MATRIX2D_H_


YG("Math");

/* @class Matrix2d,Object
This class represents a double precision 2x2 square matrix.
 */
YC class _Matrix2d : public YAC_Object {
  public:
   sF64 doubles[4];

   public:
   _Matrix2d(void);
   ~_Matrix2d();
   YAC_POOLED(_Matrix2d,YAC_POOL_PRIORITY_HIGH);
   sUI YAC_VCALL yacArrayGetNumElements(void);
   sUI YAC_VCALL yacArrayGetMaxElements(void);
   sUI YAC_VCALL yacArrayGetElementType(void);
   sUI YAC_VCALL yacArrayGetElementByteSize(void);
   void YAC_VCALL yacArraySet(YAC_ContextHandle, sUI, YAC_Value *);
   void YAC_VCALL yacArrayGet(YAC_ContextHandle, sUI, YAC_Value *);
   void *YAC_VCALL yacArrayGetPointer(void);
   sUI YAC_VCALL yacArrayGetWidth(void);
   sUI YAC_VCALL yacArrayGetHeight(void);
   sUI YAC_VCALL yacArrayGetStride(void);
   sSI YAC_VCALL yacTensorRank(void);
   void YAC_VCALL yacSerialize(YAC_Object *, sUI);
   sUI YAC_VCALL yacDeserialize(YAC_Object *, sUI);
   void YAC_VCALL yacOperatorClamp(YAC_Object *, YAC_Object *);
   void YAC_VCALL yacOperatorWrap(YAC_Object *, YAC_Object *);
//YAC_OP_CEQ, YAC_OP_CNE, YAC_OP_ASSIGN, YAC_OP_ADD = _add,
//YAC_OP_SUB = _sub, YAC_OP_MUL = _mul, YAC_OP_NEG, YAC_OP_INIT
   void YAC_VCALL yacOperator(sSI, YAC_Object *, YAC_Value *);
   void YAC_VCALL yacOperatorI(sSI, sSI, YAC_Value *);
   void YAC_VCALL yacOperatorF32(sSI, sF32, YAC_Value *);
   void YAC_VCALL yacOperatorF64(sSI, sF64, YAC_Value *);
   sUI   YAC_VCALL yacOperatorPriority       (void) { return YAC_OP_PRIO_DOUBLE; }
   sBool YAC_VCALL yacToString(YAC_String *) const;

   // Helper methods
   void assignFloats(sF32 *_fv, sUI _ne);
   void assignFloatArray(YAC_FloatArray *_fa);
   void assignFloatMatrix(YAC_Object *_o);
   void assignGeneric(YAC_ContextHandle _ctx, YAC_Object *_o);


//element access methods
   /* @method setA,Double v
      Set a element to v
      @arg v new value for a
    */
   YM void _setA(YAC_Object *);
   /* @method setB,Double v
      Set b element to v
      @arg v new value for b
    */
   YM void _setB(YAC_Object *);
   /* @method setC,Double v
      Set c element to v
      @arg v new value for c
    */
   YM void _setC(YAC_Object *);
   /* @method setD,Double v
      Set d element to v
      @arg v new value for d
    */
   YM void _setD(YAC_Object *);

   /* @method getA:Double
      Get value of a element
      @return value of a element
    */
   YM void _getA_YAC_RVAL(YAC_Value *);
   YM void _getA_YAC_RARG(YAC_Object *);
   /* @method getB:Double
      Get value of b element
      @return value of b element
    */
   YM void _getB_YAC_RVAL(YAC_Value *);
   YM void _getB_YAC_RARG(YAC_Object *);
   /* @method getC:Double
      Get value of c element
      @return value of c element
    */
   YM void _getC_YAC_RVAL(YAC_Value *);
   YM void _getC_YAC_RARG(YAC_Object *);
   /* @method getD:Double
      Get value of d element
      @return value of d element
    */
   YM void _getD_YAC_RVAL(YAC_Value *);
   YM void _getD_YAC_RARG(YAC_Object *);

//produce string representation
   /* @method getString:String
      Get string representation of matrix ((a, b), (c, d))
      @return string representation of matrix
    */
   YM void _getString(YAC_Value *) const;

//arithmetics
   /* @method add,Matrix2d m
      Add another matrix m to this instance
      @arg m matrix to add
    */
   YM void _add_YAC_RSELF(YAC_Object *);
   YM void _add_YAC_RVAL(YAC_Object *, YAC_Value *);
   YM void _add_YAC_RARG(YAC_Object *, YAC_Object *);

   /* @method sub,Matrix2d m
      Substract another matrix m from this instance
      @arg m matrix to substract
    */
   YM void _sub_YAC_RSELF(YAC_Object *);
   YM void _sub_YAC_RVAL(YAC_Object *, YAC_Value *);
   YM void _sub_YAC_RARG(YAC_Object *, YAC_Object *);

//multiply with scalar
   /* @method muld,Double v
      Multiply this instance with scalar value v
      @arg v value to multiply
    */
   void muld(sF64);
   YM void _muld_YAC_RSELF(YAC_Object *);
   YM void _muld_YAC_RVAL(YAC_Object *, YAC_Value *);
   YM void _muld_YAC_RARG(YAC_Object *, YAC_Object *);

//matrix multiplication
   /* @method mul,Matrix2d m
      Multiply this instance with matrix m
      @arg m matrix to multiply
    */
   YM void _mul_YAC_RSELF(YAC_Object *);
   YM void _mul_YAC_RVAL(YAC_Object *, YAC_Value *);
   YM void _mul_YAC_RARG(YAC_Object *, YAC_Object *);

//multiply matrix with vector (result is a vector)
   /* @method mulv,Vector2d v:Vector2d
      Multiply this instance with a 2dim vector
      @arg v vector to multiply
      @return resulting vector
    */
   YM void _mulv_YAC_RVAL(YAC_Object *, YAC_Value *);
   YM void _mulv_YAC_RARG(YAC_Object *, YAC_Object *);

//get "absolute" value for matrix
   /* @method getAbs:Double
      Calculate the Hilbert-Schmidt norm of the matrix
      @return scalar matrix norm value
    */
   sF64 getAbs(void);
   YM void _getAbs_YAC_RVAL(YAC_Value *);
   YM void _getAbs_YAC_RARG(YAC_Object *);

   /* @method getAbsSqr:Double
      Calculate the squared Hilbert-Schmidt norm of the matrix
      @return scalar matrix norm value
    */
   sF64 getAbsSqr(void);
   YM void _getAbsSqr_YAC_RVAL(YAC_Value *);
   YM void _getAbsSqr_YAC_RARG(YAC_Object *);

//normalize matrix

   /* @method unit
      Normalise matrix with Hilbert-Schmidt norm
    */
   YM void _unit_YAC_RSELF(void); /* unit */
   YM void _unit_YAC_RVAL(YAC_Value *);
   YM void _unit_YAC_RARG(YAC_Object *);

//normalize and scale matrix
   /* @method unitScale,Double v
      Normalise matrix with Hilbert-Schmidt norm and multiply with v afterwards
      @arg v value to scale matrix
    */
   void unitScale(sF64);
   YM void _unitScale_YAC_RSELF(YAC_Object *);
   YM void _unitScale_YAC_RVAL(YAC_Object *, YAC_Value *);
   YM void _unitScale_YAC_RARG(YAC_Object *, YAC_Object *);

//get scalar determinabt
   /* @method det:Double
      Calculate determinant of matrix
      @return value of determinant
    */
   sF64 det(void);
   YM void _det_YAC_RVAL(YAC_Value *);
   YM void _det_YAC_RARG(YAC_Object *);

//invert matrix so the A*A^-1=E
   /* @method invert
      Calculate inversion of matrix, so that A*A<sup>-1</sup>=U
    */
   YM void _invert_YAC_RSELF(void);
   YM void _invert_YAC_RVAL(YAC_Value *);
   YM void _invert_YAC_RARG(YAC_Object *);

//transpose matrix
   /* @method transpose
      Transpose matrix
    */
   YM void _transpose_YAC_RSELF(void);
   YM void _transpose_YAC_RVAL(YAC_Value *);
   YM void _transpose_YAC_RARG(YAC_Object *);

//access row and column vectors
   /* @method setRow,int r,Vector2d v
      Set row r with values from v
      @arg r row number
      @arg v Vector2f instance
    */
   YM void _setRow(sUI, YAC_Object*);

   /* @method getRow,int r:Vector2d
      Get row r values
      @arg r row number
      @return vector with values from row r
    */
   YM void _getRow_YAC_RVAL(sUI, YAC_Value*);
   YM void _getRow_YAC_RARG(sUI, YAC_Object*);

   /* @method setCol,int c,Vector2d v
      Set column c with values from v
      @arg c column number
      @arg v Vector2f instance
    */
   YM void _setCol(sUI, YAC_Object*);

   /* @method getCol,int c:Vector2d
      Get column c values
      @arg c column number
      @return vector with values from column c
    */
   YM void _getCol_YAC_RVAL(sUI, YAC_Value*);
   YM void _getCol_YAC_RARG(sUI, YAC_Object*);
//init matrix values
   /* @method init,Doubla va,Double vb,Double vc,Double vd
      Initialise matrix
      @arg va value for a
      @arg vb value for b
      @arg vc value for c
      @arg vd value for d
    */
   YM void _init(YAC_Object *, YAC_Object *, YAC_Object *, YAC_Object *);
   /* @method initf,float va,float vb,float vc,float vd
      Initialise matrix
      @arg va value for a
      @arg vb value for b
      @arg vc value for c
      @arg vd value for d
    */
   YM void _initf(sF32,sF32,sF32,sF32);
   /* @method initRow,int r,Double va,Double vb
      Initialise row r of matrix
      @arg r row number
      @arg va value for first element
      @arg vb value for second element
    */
   YM void _initRow(sUI,YAC_Object *, YAC_Object *);
   /* @method initCol,int c,Double va,Double vb
      Initialise column c of matrix
      @arg c column number
      @arg va value for first element
      @arg vb value for second element
    */
   YM void _initCol(sUI,YAC_Object *, YAC_Object *);
   /* @method initRowV,int r,Vector2d v
      Initialise row r of matrix
      @arg r row number
      @arg v vector with values
    */
   YM void _initRowV(sUI,YAC_Object*);
   /* @method initColV,int c,Vector2d v
      Initialise column c of matrix
      @arg c column number
      @arg v vector with values
    */
   YM void _initColV(sUI,YAC_Object*);
   /* @method New,Double va,Double vb,Double vc,Double vd:Matrix2d
      Initialise and return new instance of Matrix2d
      @arg va value for a
      @arg vb value for b
      @arg vc value for c
      @arg vd value for d
      @return new instance of Matrix2d
    */
   YM void _New(YAC_Object *, YAC_Object *, YAC_Object *, YAC_Object *, YAC_Value *);


   /* @method initIdentity
      Initialize with identity matrix.
   */
   YM void _initIdentity(void);
};

/* @function matrix2d,Double va,Double vb,Double vc,Double vd:Matrix2d
   Initialise and return new instance of Matrix2d
   @arg va value for a
   @arg vb value for b
   @arg vc value for c
   @arg vd value for d
   @return new instance of Matrix2d
 */
YF void YAC_CALL _matrix2d(YAC_Object *, YAC_Object *, YAC_Object *, YAC_Object *,YAC_Value*);
#endif
