/// matrix3f.h
///
/// (c) 2008-2020 by Carsten Busse <carsten.busse@googlemail.com>
///                  Bastian Spiegel <bs@tkscript.de> (additional coding)
///     - Distributed under terms of the Lesser GNU General Public License (LGPL).
///       See COPYING and <http://www.gnu.org/licenses/licenses.html#LGPL> for further information.
///
///

#ifndef __MATRIX3F_H_
#define __MATRIX3F_H_


YG("Math");

/* @class Matrix3f,Object
This class represents a single precision 3x3 square matrix.
 */
YC class _Matrix3f : public _Matrix {
  public:
   sF32 floats[9];

  public:
   _Matrix3f(void);
   ~_Matrix3f();
   YAC_POOLED(_Matrix3f,YAC_POOL_PRIORITY_HIGH);
   sUI YAC_VCALL yacArrayGetNumElements(void);
   sUI YAC_VCALL yacArrayGetMaxElements(void);
   sUI YAC_VCALL yacArrayGetElementType(void);
   sUI YAC_VCALL yacArrayGetElementByteSize(void);
   void YAC_VCALL yacArraySet(YAC_ContextHandle, sUI, YAC_Value *);
   void YAC_VCALL yacArrayGet(YAC_ContextHandle, sUI, YAC_Value *);
   void *YAC_VCALL yacArrayGetPointer(void);
   sUI YAC_VCALL yacArrayGetWidth(void);
   sUI YAC_VCALL yacArrayGetHeight(void);
   sUI YAC_VCALL yacArrayGetStride(void);
   sSI YAC_VCALL yacTensorRank(void);
   void YAC_VCALL yacSerialize(YAC_Object *, sUI);
   sUI YAC_VCALL yacDeserialize(YAC_Object *, sUI);
   void YAC_VCALL yacOperatorClamp(YAC_Object *, YAC_Object *);
   void YAC_VCALL yacOperatorWrap(YAC_Object *, YAC_Object *);
//YAC_OP_CEQ, YAC_OP_CNE, YAC_OP_ASSIGN, YAC_OP_ADD = _add,
//YAC_OP_SUB = _sub, YAC_OP_MUL = _mul, YAC_OP_NEG, YAC_OP_INIT
   void YAC_VCALL yacOperator(sSI, YAC_Object *, YAC_Value *);
   void YAC_VCALL yacOperatorI(sSI, sSI, YAC_Value *);
   void YAC_VCALL yacOperatorF32(sSI, sF32, YAC_Value *);
   sUI   YAC_VCALL yacOperatorPriority       (void) { return YAC_OP_PRIO_FLOAT; }
   sBool YAC_VCALL yacToString(YAC_String*) const;

   // Helper methods
   void assignFloats(sF32 *_fv, sUI _ne);
   void assignFloatArray(YAC_FloatArray *_fa);
   void assignGeneric(YAC_ContextHandle _ctx, YAC_Object *_o);

//element access methods
   /* @method setA,float v
      Set a element to v
      @arg v new value for a
    */
   YM void _setA(sF32);
   /* @method setB,float v
      Set b element to v
      @arg v new value for b
    */
   YM void _setB(sF32);
   /* @method setC,float v
      Set c element to v
      @arg v new value for c
    */
   YM void _setC(sF32);
   /* @method setD,float v
      Set d element to v
      @arg v new value for d
    */
   YM void _setD(sF32);
   /* @method setE,float v
      Set e element to v
      @arg v new value for e
    */
   YM void _setE(sF32);
   /* @method setF,float v
      Set f element to v
      @arg v new value for f
    */
   YM void _setF(sF32);
   /* @method setG,float v
      Set g element to v
      @arg v new value for g
    */
   YM void _setG(sF32);
   /* @method setH,float v
      Set h element to v
      @arg v new value for h
    */
   YM void _setH(sF32);
   /* @method setI,float v
      Set i element to v
      @arg v new value for i
    */
   YM void _setI(sF32);

   /* @method getA:float
      Get value of a element
      @return value of a element
    */
   YM sF32 _getA(void);
   /* @method getB:float
      Get value of b element
      @return value of b element
    */
   YM sF32 _getB(void);
   /* @method getC:float
      Get value of c element
      @return value of c element
    */
   YM sF32 _getC(void);
   /* @method getD:float
      Get value of d element
      @return value of d element
    */
   YM sF32 _getD(void);
   /* @method getE:float
      Get value of e element
      @return value of e element
    */
   YM sF32 _getE(void);
   /* @method getF:float
      Get value of f element
      @return value of f element
    */
   YM sF32 _getF(void);
   /* @method getG:float
      Get value of g element
      @return value of g element
    */
   YM sF32 _getG(void);
   /* @method getH:float
      Get value of h element
      @return value of h element
    */
   YM sF32 _getH(void);
   /* @method getI:float
      Get value of i element
      @return value of i element
    */
   YM sF32 _getI(void);

//produce string representation
   /* @method getString:String
      Get string representation of matrix ((a, b, c), (d, e, f), (g, h, i))
      @return string representation of matrix
    */
   YM void _getString(YAC_Value *) const;

//arithmetics
   /* @method add,Matrix3f m
      Add another matrix m to this instance
      @arg m matrix to add
    */
   YM void _add_YAC_RSELF(YAC_Object *);
   YM void _add_YAC_RVAL(YAC_Object *, YAC_Value *);
   YM void _add_YAC_RARG(YAC_Object *, YAC_Object *) const;

   /* @method sub,Matrix3f m
      Substract another matrix m from this instance
      @arg m matrix to substract
    */
   YM void _sub_YAC_RSELF(YAC_Object *);
   YM void _sub_YAC_RVAL(YAC_Object *, YAC_Value *);
   YM void _sub_YAC_RARG(YAC_Object *, YAC_Object *) const;

//multiply with scalar
   /* @method mulf,float v
      Multiply this instance with scalar value v
      @arg v value to multiply
    */
   YM void _mulf_YAC_RSELF(sF32);
   YM void _mulf_YAC_RVAL(sF32, YAC_Value *);
   YM void _mulf_YAC_RARG(sF32, YAC_Object *) const;

//matrix multiplication
   /* @method mul,Matrix3f m
      Multiply this instance with matrix m
      @arg m matrix to multiply
    */
   YM void _mul_YAC_RSELF(YAC_Object *);
   YM void _mul_YAC_RVAL(YAC_Object *, YAC_Value *);
   YM void _mul_YAC_RARG(YAC_Object *, YAC_Object *) const;

//multiply matrix with vector (result is a vector)
   /* @method mulv,Vector3f v:Vector3f
      Multiply this instance with a 3dim vector
      @arg v vector to multiply
      @return resulting vector
    */
   YM void _mulv_YAC_RVAL(YAC_Object *, YAC_Value *);
   YM void _mulv_YAC_RARG(YAC_Object *, YAC_Object *) const;

//get "absolute" value for matrix
   /* @method getAbs:float
      Calculate the Hilbert-Schmidt norm of the matrix
      @return scalar matrix norm value
    */
   YM sF32 _getAbs(void) const;

   /* @method getAbsSqr:float
      Calculate the squared Hilbert-Schmidt norm of the matrix
      @return scalar matrix norm value
    */
   YM sF32 _getAbsSqr(void) const;

//normalize matrix
   /* @method unit
      Normalise matrix with Hilbert-Schmidt norm
    */
   YM void _unit_YAC_RSELF(void); /* unit */
   YM void _unit_YAC_RVAL(YAC_Value *);
   YM void _unit_YAC_RARG(YAC_Object *) const;

//normalize and scale matrix
   /* @method unitScale,float v
      Normalise matrix with Hilbert-Schmidt norm and multiply with v afterwards
      @arg v value to scale matrix
    */
   YM void _unitScale_YAC_RSELF(sF32);
   YM void _unitScale_YAC_RVAL(sF32, YAC_Value *);
   YM void _unitScale_YAC_RARG(sF32, YAC_Object *) const;

//get scalar determinant
   /* @method det:float
      Calculate determinant of matrix
      @return value of determinant
    */
   YM sF32 _det(void) const;

//invert matrix so the A*A^-1=E
   /* @method invert
      Calculate inversion of matrix, so that A*A<sup>-1</sup>=U
    */
   YM void _invert_YAC_RSELF(void);
   YM void _invert_YAC_RVAL(YAC_Value *);
   YM void _invert_YAC_RARG(YAC_Object *) const;

//transpose matrix
   /* @method transpose
      Transpose matrix
    */
   YM void _transpose_YAC_RSELF(void);
   YM void _transpose_YAC_RVAL(YAC_Value *);
   YM void _transpose_YAC_RARG(YAC_Object *) const;

//access row and column vectors
   /* @method setRow,int r,Vector3f v
      Set row r with values from v
      @arg r row number
      @arg v Vector3f instance
    */
   YM void _setRow(sUI, YAC_Object*);
   /* @method getRow,int r:Vector3f
      Get row r values
      @arg r row number
      @return vector with values from row r
    */
   YM void _getRow_YAC_RVAL(sUI, YAC_Value*);
   YM void _getRow_YAC_RARG(sUI, YAC_Object*) const;
   /* @method setCol,int c,Vector3f v
      Set column c with values from v
      @arg c column number
      @arg v Vector3f instance
    */
   YM void _setCol(sUI, YAC_Object*);
   /* @method getCol,int c:Vector3f
      Get column c values
      @arg c column number
      @return vector with values from column c
    */
   YM void _getCol_YAC_RVAL(sUI, YAC_Value*);
   YM void _getCol_YAC_RARG(sUI, YAC_Object*) const;

//init matrix values
   /* @method init,float va,float vb,float vc,float vd,float ve, float vf, float vg, float vh, float vi
      Initialise matrix
      @arg va value for a
      @arg vb value for b
      @arg vc value for c
      @arg vd value for d
      @arg ve value for e
      @arg vf value for f
      @arg vg value for g
      @arg vh value for h
      @arg vi value for i
    */
   YM void _init(sF32,sF32,sF32,sF32,sF32,sF32,sF32,sF32,sF32);
//row, x, y, z
   /* @method initRow,int r,float va,float vb,float vc
      Initialise row r of matrix
      @arg r row number
      @arg va value for first element
      @arg vb value for second element
      @arg vc value for third element
    */
   YM void _initRow(sUI,sF32,sF32,sF32);
//col, x, y, z
   /* @method initCol,int c,float va,float vb,float vc
      Initialise column c of matrix
      @arg c column number
      @arg va value for first element
      @arg vb value for second element
      @arg vc value for third element
    */
   YM void _initCol(sUI,sF32,sF32,sF32);
   /* @method initRowV,int r,Vector3f v
      Initialise row r of matrix
      @arg r row number
      @arg v vector with values
    */
   YM void _initRowV(sUI,YAC_Object*);
   /* @method initColV,int c,Vector3f v
      Initialise column c of matrix
      @arg c column number
      @arg v vector with values
    */
   YM void _initColV(sUI,YAC_Object*);

   /* @method _initScalef
    */
   YM void _initScalef (sF32 _x, sF32 _y, sF32 _z);

   /* @method _initScalev
      @arg scale Vector2f, or Vector3f
    */
   YM void _initScalev (YAC_Object *_scale);

   /* @method initEulerXYX,float a,float b,float c
      Initialise with euler rotation matrix
      @arg a first rotation
      @arg b second rotation
      @arg c third rotation
    */
   YM void _initEulerXYX(sF32, sF32, sF32);


   /* @method initEulerXZX,float a,float b,float c
      Initialise with euler rotation matrix
      @arg a first rotation
      @arg b second rotation
      @arg c third rotation
    */
   YM void _initEulerXZX(sF32, sF32, sF32);


   /* @method initEulerYXY,float a,float b,float c
      Initialise with euler rotation matrix
      @arg a first rotation
      @arg b second rotation
      @arg c third rotation
    */
   YM void _initEulerYXY(sF32, sF32, sF32);


   /* @method initEulerYZY,float a,float b,float c
      Initialise with euler rotation matrix
      @arg a first rotation
      @arg b second rotation
      @arg c third rotation
    */
   YM void _initEulerYZY(sF32, sF32, sF32);


   /* @method initEulerZXZ,float a,float b,float c
      Initialise with euler rotation matrix
      @arg a first rotation
      @arg b second rotation
      @arg c third rotation
    */
   YM void _initEulerZXZ(sF32, sF32, sF32);


   /* @method initEulerZYZ,float a,float b,float c
      Initialise with euler rotation matrix
      @arg a first rotation
      @arg b second rotation
      @arg c third rotation
    */
   YM void _initEulerZYZ(sF32, sF32, sF32);


   /* @method initEulerXYZ,float a,float b,float c
      Initialise with euler rotation matrix
      @arg a first rotation
      @arg b second rotation
      @arg c third rotation
    */
   YM void _initEulerXYZ(sF32, sF32, sF32);


   /* @method initEulerXZY,float a,float b,float c
      Initialise with euler rotation matrix
      @arg a first rotation
      @arg b second rotation
      @arg c third rotation
    */
   YM void _initEulerXZY(sF32, sF32, sF32);


   /* @method initEulerYXZ,float a,float b,float c
      Initialise with euler rotation matrix
      @arg a first rotation
      @arg b second rotation
      @arg c third rotation
    */
   YM void _initEulerYXZ(sF32, sF32, sF32);


   /* @method initEulerYZX,float a,float b,float c
      Initialise with euler rotation matrix
      @arg a first rotation
      @arg b second rotation
      @arg c third rotation
    */
   YM void _initEulerYZX(sF32, sF32, sF32);


   /* @method initEulerZXY,float a,float b,float c
      Initialise with euler rotation matrix
      @arg a first rotation
      @arg b second rotation
      @arg c third rotation
    */
   YM void _initEulerZXY(sF32, sF32, sF32);


   /* @method initEulerZYX,float a,float b,float c
      Initialise with euler rotation matrix
      @arg a first rotation
      @arg b second rotation
      @arg c third rotation
    */
   YM void _initEulerZYX(sF32, sF32, sF32);


   /* @method New,float va,float vb,float vc,float vd,float ve,float vf,float vg,float vh,float vi:Matrix3f
      Initialise and return new instance of Matrix3f
      @arg va value for a
      @arg vb value for b
      @arg vc value for c
      @arg vd value for d
      @arg ve value for e
      @arg vf value for f
      @arg vg value for g
      @arg vh value for h
      @arg vi value for i
      @return new instance of Matrix3f
    */
   YM void _New(sF32,sF32,sF32,sF32,sF32,sF32,sF32,sF32,sF32,YAC_Value*);


   /* @method initIdentity
      Initialize with identity matrix.
   */
   YM void _initIdentity(void);
};

/* @function matrix3f,float va,float vb,float vc,float vd,float ve,float vf,float vg,float vh,float vi:Matrix3f
   Initialise and return new instance of Matrix3f
   @arg va value for a
   @arg vb value for b
   @arg vc value for c
   @arg vd value for d
   @arg ve value for e
   @arg vf value for f
   @arg vg value for g
   @arg vh value for h
   @arg vi value for i
   @return new instance of Matrix3f
 */
YF void YAC_CALL _matrix3f(sF32,sF32,sF32,sF32,sF32,sF32,sF32,sF32,sF32,YAC_Value*);
#endif
